/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */
#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/cdefs.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <sys/proxy.h>
#include <sys/prfx.h>
#include <sys/psinfo.h>
#include <sys/stat.h>
#include <sys/io_msg.h>
#include <sys/fsys.h>
#include <sys/fd.h>

#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <unix.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/route.h>
#include <net/if_vp.h>
#include <netinet/ip.h>

#include "netinet/ip_compat.h"
#include "netinet/ip_fil.h"
#include "netinet/ip_nat.h"
#include "netinet/ip_state.h"
#include "ipf_res.h"

static void *fd_ctrl;
static char *ipl_name[NUM_DEVS] = {
	IPL_NAME,
	IPL_NAT,
	IPL_STATE,
#ifdef IPFILTER_AUTH
	IPL_AUTH
#endif
};

#define NUM_READERS 10

/* globals */
struct stat	ipl_stat[NUM_DEVS];
int    ipl_devno[NUM_DEVS];
union  ipl_msg iomsg;
struct reader_list_t reader_list[NUM_DEVS];

extern filterstats_t frstats[];

static mode_t check_access(pid_t pid, struct stat *sptr)
{
	mode_t  mode = 0;
	uid_t   uid;
	gid_t   gid;
	struct _psinfo2  psinfo2;
	static const mode_t perm_mode[3] = {FREAD, FWRITE, FREAD|FWRITE};

	/* Get egid and euid of process pid.*/
	if (qnx_getids( pid, &psinfo2) == -1)
	  return mode;
	uid = psinfo2.euid;
    gid = psinfo2.egid;

	/* Make sure the calling process has correct permissions. */
	mode = perm_mode[iomsg.open.oflag & O_ACCMODE];
	mode &= (uid == 0) ?
	  S_IRWXU : (uid == sptr->st_uid) ? (sptr->st_mode & S_IRWXU) :
	(gid == sptr->st_gid) ? (sptr->st_mode & S_IRWXG) << 3 :
        (sptr->st_mode & S_IRWXO) << 6;
	return(mode);
}

static struct ocb * get_ocb( pid_t pid, int fd)
{
	struct ocb      *ocb;

	ocb = (struct ocb *)__get_fd(pid, fd, fd_ctrl);
	if (ocb == (struct ocb *)-1 || ocb == NULL) 
	{
		iomsg.status = EBADF;
		return NULL;
	}
	return(ocb);
}

void ipf_res_cleanup()
{
	int i;
	
	for (i = 0; i < NUM_DEVS; i++) {
		qnx_prefix_detach(ipl_name[i]);
		qnx_device_detach(ipl_devno[i]);
	}

	return;
}

int ipf_res_init()
{
	time_t t;
	int i;

	if (qnx_pflags(~0, _PPF_SERVER | _PPF_PRIORITY_REC | _PPF_PRIORITY_FLOAT | _PPF_SIGCATCH, NULL, NULL))
	{
		perror("qnx_pflags");
		return -1;
	}

	if (!(fd_ctrl = __init_fd(getpid()))) {
		perror("__init_fd");
		return -1;
	}

	/* Attach device numbers to our I/O manager */
	for (i = 0; i < NUM_DEVS; i++) {
		if ((ipl_devno[i] = qnx_device_attach()) == -1)
		{
			perror("qnx_device_attach");
			return -1;
		}

		if (qnx_prefix_attach(ipl_name[i], NULL, ipl_devno[i]) != 0)
		{
			perror("qnx_prefix_attach");
			return -1;
		}

		/* Initialize individual stat structure. */
		/* file serial number */
		ipl_stat[i].st_ino = i + 1;
		ipl_stat[i].st_rdev = (dev_t) ((getnid() << 16L) + (ipl_devno[i] << 10L));
		ipl_stat[i].st_mode = S_IFCHR | S_IRUSR | S_IWUSR;
		ipl_stat[i].st_nlink = 1;           /* number of links.				*/
		ipl_stat[i].st_ftime = (t= time(0));/* Time of creation.				*/
		ipl_stat[i].st_atime = t;		 /* Time of last access.			*/
		ipl_stat[i].st_mtime = t;           /* Time of last modification.	*/
		ipl_stat[i].st_ctime = t;           /* Time of last status change.	*/
		ipl_stat[i].st_uid = 0;
		ipl_stat[i].st_gid = 0;
		ipl_stat[i].st_status = _FILE_USED;
		ipl_stat[i].st_size = 0;

		/* init the reader list for this device */
		if ((reader_list[i].readers = calloc(sizeof(struct reader), NUM_READERS)) != NULL) {
			reader_list[i].head = reader_list[i].tail = 0;
			reader_list[i].maxlen = NUM_READERS;
		}
	}

	return 0;
}

/*
 * routines below for saving IP headers to buffer
 */
int ipf_open(pid_t pid)
{
	mode_t          mode;
	struct ocb      *ocb;
	int             devid;

	for (devid = 0; devid < NUM_DEVS; devid++)
	  if (iomsg.open.unit == ipl_devno[devid])
		break;

	if (devid == NUM_DEVS) {
		iomsg.status = EBADF;
		return sizeof(iomsg.status);
	}

	if(((mode=check_access(pid, &ipl_stat[devid])) & S_IRWXU) == 0)
	{
		iomsg.status = EPERM;
		return sizeof(iomsg.status);
	}

	if ((ocb = (struct ocb *)malloc(sizeof(*ocb)))== NULL) {
		iomsg.status = errno;
		return sizeof(iomsg.status);
	}

    if(qnx_fd_attach(pid, iomsg.open.fd, 0, 0, 0, 0, (unsigned)ocb) == -1)
	{
		iomsg.status = errno;
		return sizeof(iomsg.status);
	}

	ocb->mode = mode; /* Save access mode bits. */
	ocb->unit = devid;
	ocb->rd   = NULL;
	iomsg.open_reply.status = EOK;
	return sizeof(iomsg.open_reply);
}

int ipf_close(pid_t pid, struct ocb *ocb)
{
	qnx_fd_attach(pid, iomsg.close.fd, 0, 0, 0, 0, (unsigned)NULL);
	qnx_fd_detach(iomsg.close.fd);

	if (ocb->rd)
	  ocb->rd->pid = 0;

	free(ocb);

	iomsg.close_reply.status = EOK;
	return sizeof(iomsg.close_reply);
}

int ipf_fstat(pid_t pid, struct ocb *ocb)
{
	memcpy(&iomsg.fstat_reply.stat, &ipl_stat[ocb->unit], sizeof(struct stat));
	iomsg.fstat_reply.status = EOK;
	
	return sizeof(iomsg.fstat_reply);
}

int ipf_iostat(pid_t pid)
{
	int             devid;

	for (devid = 0; devid < NUM_DEVS; devid++)
	  if (iomsg.open.unit == ipl_devno[devid])
		break;

	if(devid >= NUM_DEVS || iomsg.open.path[0]  == NULL) {
		memcpy(&iomsg.fstat_reply.stat, &ipl_stat[devid], sizeof(struct stat));
		iomsg.fstat_reply.status = EOK;
	} else {
		iomsg.status = ENOENT;
		return sizeof(iomsg.fstat_reply);
	}
	iomsg.status = EOK;
	return sizeof(iomsg.fstat_reply);
}

int ipf_sioctl(pid_t pid, struct ocb *ocb)
{
	switch(ocb->unit) {
	  case DEV_IPL:
		iomsg.ioctl_reply.status = ipf_ioctl(&iomsg.ioctl.data, iomsg.ioctl.zero, ocb->mode, pid);
		break;
	  case DEV_IPNAT:
		iomsg.ioctl_reply.status = nat_ioctl(&iomsg.ioctl.data, iomsg.ioctl.zero, ocb->mode, pid);
		break;
	  case DEV_IPSTATE:
		iomsg.ioctl_reply.status = fr_state_ioctl(&iomsg.ioctl.data, iomsg.ioctl.zero, ocb->mode, pid);
		break;
#ifdef IPFILTER_AUTH
	  case DEV_IPAUTH:
		iomsg.ioctl_reply.status = fr_auth_ioctl(&iomsg.ioctl.data, iomsg.ioctl.zero, NULL, NULL, pid);
		break;
#endif
	}

	return offsetof(struct _io_ioctl_reply, data);
}

int ipf_read(pid_t pid, struct ocb *ocb)
{
#ifdef IPFILTER_LOG
	int ret;
	struct reader_list_t *rlist;
	struct reader *r;
	
	ret = ipflog_read(pid, ocb->unit, &iomsg.read);
	switch (ret) {
	  case -1:
		/* success, do the message reply */
		iomsg.read_reply.status = errno;
		break;
	  case 0:
		/* no data, block reader */
		rlist = &reader_list[ocb->unit];

		/* reader list is not ready */
		if (!rlist->readers) {
			iomsg.read_reply.status = ENXIO;
			break;
		}
		
		/* reader list is full */
		if ((rlist->tail + 1 == rlist->head) || (rlist->tail + 1 == rlist->maxlen && rlist->head == 0)) {
			iomsg.read_reply.status = ENXIO;
			break;
		}

		if (++rlist->tail == rlist->maxlen)
		  rlist->tail = 0;

		r = &rlist->readers[rlist->tail];
		r->ocb = ocb;
		r->pid = pid;
		r->msg = iomsg.read;
		ocb->rd = r;
		return NO_REPLY;

	  default:
		/* succeesed, return real bytes */
		iomsg.read_reply.status = EOK;
		iomsg.read_reply.nbytes = ret;
		break;
	}
# else
	iomsg.read_reply.status = ENXIO;
# endif

 	return offsetof(struct _io_read_reply, data);
}

void ipf_res_intr(pid_t pid)
{
	struct ocb      *ocb;
	int ret;

	switch(iomsg.type) {
	  case _IO_OPEN:
		ret = ipf_open(pid);
		break;
	  case _IO_READ:
		if ((ocb = get_ocb(pid, iomsg.read.fd)) == (struct ocb *)-1)
		  ret = sizeof(iomsg.status);
		else
		  ret = ipf_read(pid, ocb);
		break;
	  case _IO_CLOSE:
		if ((ocb = get_ocb(pid, iomsg.close.fd)) == (struct ocb *)-1)
		  ret = sizeof(iomsg.status);
		else
		  ret = ipf_close(pid, ocb);
		break;
	  case _IO_IOCTL:
	  case _IO_QIOCTL:
		if ((ocb = get_ocb(pid, iomsg.ioctl.fd)) == (struct ocb *)-1)
		  ret = sizeof(iomsg.status);
		else
		  ret = ipf_sioctl(pid, ocb);
		break;
	  case _IO_STAT:
		ret = ipf_iostat(pid);
		break;
	  case _IO_FSTAT:
		if ((ocb = get_ocb(pid, iomsg.fstat.fd)) == (struct ocb *)-1)
		  ret = sizeof(iomsg.status);
		else
		  ret = ipf_fstat(pid, ocb);
		break;

	  case _SYSMSG:
		if (iomsg.sysmsg.hdr.subtype == _SYSMSG_SUBTYPE_MEMCPY) {
			struct _mxfer_entry mx[2];

			iomsg.sysmsg_reply.hdr_reply.status = EOK;
			_setmx(&mx[0], &iomsg, sizeof(iomsg.sysmsg_reply));
			_setmx(&mx[1], iomsg.sysmsg.copy.from, iomsg.sysmsg.copy.size);
			Replymx(pid, 2, mx);
			ret = 0;
		} else {
			iomsg.status = ENOSYS;
			ret = sizeof(iomsg.status);
		}
		break;
	  default:
		iomsg.status = ENOSYS;
		ret = sizeof(iomsg.status);
		break;
	}

    if (ret != 0)
	  Reply(pid, &iomsg, ret);
	return;
}
